/*
 * Copyright (C) 1998 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file mixer.c A volume fader and alarm clock for OSS. */

#include <regex.h>
#include <lopsub.h>
#include <math.h>

#include "mixer.lsg.h"
#include "para.h"
#include "fd.h"
#include "string.h"
#include "mix.h"
#include "error.h"
#include "version.h"

/** Array of error strings. */
DEFINE_PARA_ERRLIST;

/* At least one of the two is defined if this file gets compiled. */
static const struct mixer *mixers[] = {
#ifdef HAVE_ALSA
	&alsa_mixer,
#endif
#ifdef HAVE_OSS
	&oss_mixer,
#endif
};

#define NUM_SUPPORTED_MIXERS (ARRAY_SIZE(mixers))
#define FOR_EACH_MIXER(i) for ((i) = 0; (i) < NUM_SUPPORTED_MIXERS; (i)++)

static struct lls_parse_result *lpr, *sub_lpr;

#define CMD_PTR(_cmd) (lls_cmd(LSG_MIXER_CMD_ ## _cmd, mixer_suite))
#define OPT_RESULT(_cmd, _opt) (lls_opt_result( \
	LSG_MIXER_ ## _cmd ## _OPT_ ## _opt, (LSG_MIXER_CMD_ ## _cmd == 0)? lpr : sub_lpr))
#define OPT_GIVEN(_cmd, _opt) (lls_opt_given(OPT_RESULT(_cmd, _opt)))
#define OPT_STRING_VAL(_cmd, _opt) (lls_string_val(0, OPT_RESULT(_cmd, _opt)))
#define OPT_UINT32_VAL(_cmd, _opt) (lls_uint32_val(0, OPT_RESULT(_cmd, _opt)))

typedef int (*mixer_subcommand_handler_t)(const struct mixer *, struct mixer_handle *);

#define EXPORT_CMD(_cmd) const mixer_subcommand_handler_t \
	lsg_mixer_com_ ## _cmd ## _user_data = &com_ ## _cmd;

static int loglevel;
static __printf_2_3 void date_log(int ll, const char *fmt, ...)
{
	va_list argp;
	time_t t1;
	struct tm *tm;

	if (ll < loglevel)
		return;
	time(&t1);
	tm = localtime(&t1);
	fprintf(stderr, "%d:%02d:%02d ", tm->tm_hour, tm->tm_min, tm->tm_sec);
	va_start(argp, fmt);
	vprintf(fmt, argp);
	va_end(argp);
}
__printf_2_3 void (*para_log)(int, const char*, ...) = date_log;

static int set_channel(const struct mixer *m, struct mixer_handle *h,
		const char *channel)
{

	PARA_NOTICE_LOG("using %s mixer channel\n", channel?
		channel : "default");
	return m->set_channel(h, channel);
}

static void millisleep(int ms)
{
	struct timespec ts;

	PARA_INFO_LOG("sleeping %dms\n", ms);
	if (ms < 0)
		return;
	ts.tv_sec = ms / 1000,
	ts.tv_nsec = (ms % 1000) * 1000 * 1000;
	nanosleep(&ts, NULL);
}

/*
 * This implements the inverse function of t -> t^alpha, scaled to the time
 * interval [0,T] and the range given by old_vol and new_vol. It returns the
 * amount of milliseconds until the given volume is reached.
 */
static unsigned volume_time(double vol, double old_vol, double new_vol,
		double T, double alpha)
{
	double c, d, x;

	if (old_vol < new_vol) {
		c = old_vol;
		d = new_vol;
	} else {
		c = new_vol;
		d = old_vol;
	}

	x = T * exp(log(((vol - c) / (d - c))) / alpha);
	assert(x <= T);
	if (old_vol < new_vol)
		return x;
	else
		return T - x;
}

/* Fade to new volume in fade_time seconds. */
static int fade(const struct mixer *m, struct mixer_handle *h, uint32_t new_vol,
		uint32_t fade_time)
{
	int i, T, old_vol, ret, slept, incr;
	double ms, alpha;
	uint32_t fe = OPT_UINT32_VAL(PARA_MIXER, FADE_EXPONENT);

	if (fade_time <= 0 || fe >= 100) {
		ret = m->set(h, new_vol);
		if (ret < 0)
			return ret;
		goto sleep;
	}
	alpha = (100 - fe) / 100.0;
	ret = m->get(h);
	if (ret < 0)
		return ret;
	old_vol = ret;
	if (old_vol == new_vol)
		goto sleep;
	PARA_NOTICE_LOG("fading %s from %d to %u in %u seconds\n",
		OPT_STRING_VAL(PARA_MIXER, MIXER_CHANNEL), old_vol,
		new_vol, fade_time);
	incr = old_vol < new_vol? 1 : -1;
	T = fade_time * 1000;
	i = old_vol;
	slept = 0;
	do {
		ms = volume_time(i + incr, old_vol, new_vol, T, alpha);
		millisleep(ms - slept);
		slept = ms;
		i += incr;
		ret = m->set(h, i);
		if (ret < 0)
			return ret;
	} while (i != new_vol);
	return 1;
sleep:
	sleep(fade_time);
	return ret;
}

static int com_fade(const struct mixer *m, struct mixer_handle *h)
{
	uint32_t new_vol = OPT_UINT32_VAL(FADE, FADE_VOL);
	uint32_t fade_time = OPT_UINT32_VAL(FADE, FADE_TIME);
	return fade(m, h, new_vol, fade_time);
}
EXPORT_CMD(fade);

static void client_cmd(const char *cmd)
{
	int ret, status, fds[3] = {0, 0, 0};
	pid_t pid;
	char *cmdline = make_message(BINDIR "/para_client %s", cmd);

	PARA_NOTICE_LOG("%s\n", cmdline);
	ret = para_exec_cmdline_pid(&pid, cmdline, fds);
	free(cmdline);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		goto fail;
	}
	do
		pid = waitpid(pid, &status, 0);
	while (pid == -1 && errno == EINTR);
	if (pid < 0) {
		PARA_ERROR_LOG("%s\n", strerror(errno));
		goto fail;
	}
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		goto fail;
	return;
fail:
	PARA_EMERG_LOG("command \"%s\" failed\n", cmd);
	exit(EXIT_FAILURE);
}

static void change_afs_mode(const char *afs_mode)
{
	char *cmd;

	client_cmd("stop");
	if (!afs_mode)
		return;
	cmd = make_message("select %s", afs_mode);
	client_cmd(cmd);
	free(cmd);
}

static int set_initial_volume(const struct mixer *m, struct mixer_handle *h)
{
	int i, ret;

	for (i = 0; i < OPT_GIVEN(SLEEP, IVOL); i++) {
		const char *val = lls_string_val(i, OPT_RESULT(SLEEP, IVOL));
		char *p, *ch, *arg = para_strdup(val);
		int32_t iv;
		p = strchr(arg, ':');
		if (p) {
			*p = '\0';
			p++;
			ch = arg;
		} else {
			p = arg;
			ch = NULL;
		}
		ret = para_atoi32(p, &iv);
		if (ret < 0) {
			free(arg);
			return ret;
		}
		ret = set_channel(m, h, ch);
		if (!ch)
			ch = "default";
		if (ret < 0) {
			PARA_WARNING_LOG("ignoring channel %s\n", ch);
			ret = 0;
		} else {
			PARA_INFO_LOG("initial volume %s: %d\n", ch, iv);
			ret = m->set(h, iv);
		}
		free(arg);
		if (ret < 0)
			return ret;
	}
	return 1;
}

static int com_sleep(const struct mixer *m, struct mixer_handle *h)
{
	time_t t1, wake_time_epoch;
	unsigned int delay;
	struct tm *tm;
	int ret;
	const char *wake_time = OPT_STRING_VAL(SLEEP, WAKE_TIME);
	const char *fo_mood = OPT_STRING_VAL(SLEEP, FO_MOOD);
	const char *fi_mood = OPT_STRING_VAL(SLEEP, FI_MOOD);
	const char *sleep_mood = OPT_STRING_VAL(SLEEP, SLEEP_MOOD);
	int fit = OPT_UINT32_VAL(SLEEP, FI_TIME);
	int fot = OPT_UINT32_VAL(SLEEP, FO_TIME);
	int fiv = OPT_UINT32_VAL(SLEEP, FI_VOL);
	int fov = OPT_UINT32_VAL(SLEEP, FO_VOL);
	int32_t hour, min = 0;
	char *tmp;
	char *wt = para_strdup(wake_time + (wake_time[0] == '+'));

	/* calculate wake time */
	time(&t1);
	tmp = strchr(wt, ':');
	if (tmp) {
		*tmp = '\0';
		tmp++;
		ret = para_atoi32(tmp, &min);
		if (ret < 0) {
			free(wt);
			return ret;
		}
	}
	ret = para_atoi32(wt, &hour);
	free(wt);
	if (ret < 0)
		return ret;
	if (wake_time[0] == '+') { /* relative */
		t1 += hour * 60 * 60 + min * 60;
		tm = localtime(&t1);
	} else {
		tm = localtime(&t1);
		if (tm->tm_hour > hour || (tm->tm_hour == hour && tm->tm_min> min)) {
			t1 += 86400; /* wake time is tomorrow */
			tm = localtime(&t1);
		}
		tm->tm_hour = hour;
		tm->tm_min = min;
		tm->tm_sec = 0;
	}
	wake_time_epoch = mktime(tm);
	PARA_INFO_LOG("waketime: %d:%02d\n", tm->tm_hour, tm->tm_min);
	client_cmd("stop");
	sleep(1);
	if (fot) {
		ret = set_initial_volume(m, h);
		if (ret < 0)
			return ret;
		change_afs_mode(fo_mood);
		client_cmd("play");
		ret = set_channel(m, h, OPT_STRING_VAL(PARA_MIXER, MIXER_CHANNEL));
		if (ret < 0)
			return ret;
		ret = fade(m, h, fov, fot);
		if (ret < 0)
			return ret;
	} else {
		ret = m->set(h, fov);
		if (ret < 0)
			return ret;
	}
	if (OPT_GIVEN(SLEEP, SLEEP_MOOD)) {
		change_afs_mode(sleep_mood);
		client_cmd("play");
	} else
		client_cmd("stop");
	if (!fit)
		return 1;
	change_afs_mode(fi_mood);
	for (;;) {
		time(&t1);
		if (wake_time_epoch <= t1 + fit)
			break;
		delay = wake_time_epoch - t1 - fit;
		PARA_INFO_LOG("sleeping %u seconds (%u:%02u)\n",
			delay, delay / 3600,
			(delay % 3600) / 60);
		sleep(delay);
	}
	client_cmd("play");
	ret = fade(m, h, fiv, fit);
	PARA_INFO_LOG("fade complete, returning\n");
	return ret;
}
EXPORT_CMD(sleep);

static int com_snooze(const struct mixer *m, struct mixer_handle *h)
{
	int ret, val;

	if (OPT_UINT32_VAL(SNOOZE, SO_TIME) == 0)
		return 1;
	ret = m->get(h);
	if (ret < 0)
		return ret;
	val = ret;
	if (val < OPT_UINT32_VAL(SNOOZE, SO_VOL))
		ret = m->set(h, OPT_UINT32_VAL(SNOOZE, SO_VOL));
	else
		ret = fade(m, h, OPT_UINT32_VAL(SNOOZE, SO_VOL),
			OPT_UINT32_VAL(SNOOZE, SO_TIME));
	if (ret < 0)
		return ret;
	client_cmd("pause");
	PARA_NOTICE_LOG("%" PRIu32 " seconds snooze time...\n",
		OPT_UINT32_VAL(SNOOZE, SNOOZE_TIME));
	sleep(OPT_UINT32_VAL(SNOOZE, SNOOZE_TIME));
	client_cmd("play");
	return fade(m, h, OPT_UINT32_VAL(SNOOZE, SI_VOL),
		OPT_UINT32_VAL(SNOOZE, SI_TIME));
}
EXPORT_CMD(snooze);

static int com_set(const struct mixer *m, struct mixer_handle *h)
{
	return m->set(h, OPT_UINT32_VAL(SET, VAL));
}
EXPORT_CMD(set);

static const struct mixer *get_mixer_or_die(void)
{
	int i;

	if (!OPT_GIVEN(PARA_MIXER, MIXER_API))
		i = 0; /* default: use first mixer */
	else
		FOR_EACH_MIXER(i)
			if (!strcmp(mixers[i]->name,
					OPT_STRING_VAL(PARA_MIXER, MIXER_API)))
				break;
	if (i < NUM_SUPPORTED_MIXERS) {
		PARA_NOTICE_LOG("using %s mixer API\n", mixers[i]->name);
		return mixers[i];
	}
	printf("available mixer APIs: ");
	FOR_EACH_MIXER(i)
		printf("%s%s%s ", i == 0? "[" : "", mixers[i]->name,
			i == 0? "]" : "");
	printf("\n");
	exit(EXIT_FAILURE);
}

static void show_subcommands(void)
{
	const struct lls_command *cmd;
	int i;
	printf("Subcommands:\n");
	for (i = 1; (cmd = lls_cmd(i, mixer_suite)); i++) {
		const char *name = lls_command_name(cmd);
		const char *purpose = lls_purpose(cmd);
		printf("%-20s%s\n", name, purpose);
	}
}


static int com_help(__a_unused const struct mixer *m,
		__a_unused struct mixer_handle *h)
{
	const struct lls_command *cmd;
	const struct lls_opt_result *r_l = OPT_RESULT(HELP, LONG);
	char *txt, *errctx;
	const char *name;
	int ret;

	ret = lls_check_arg_count(sub_lpr, 0, 1, NULL);
	if (ret < 0)
		return ret;
	if (lls_num_inputs(sub_lpr) == 0) {
		show_subcommands();
		return 0;
	}
	name = lls_input(0, sub_lpr);
	ret = lls(lls_lookup_subcmd(name, mixer_suite, &errctx));
	if (ret < 0) {
		if (errctx)
			PARA_ERROR_LOG("%s\n", errctx);
		free(errctx);
		return ret;
	}
	cmd = lls_cmd(ret, mixer_suite);
	if (lls_opt_given(r_l))
		txt = lls_long_help(cmd);
	else
		txt = lls_short_help(cmd);
	printf("%s", txt);
	free(txt);
	return 0;
}
EXPORT_CMD(help);

static void handle_help_flags(void)
{
	char *help;

	if (OPT_GIVEN(PARA_MIXER, DETAILED_HELP))
		help = lls_long_help(CMD_PTR(PARA_MIXER));
	else if (OPT_GIVEN(PARA_MIXER, HELP))
		help = lls_short_help(CMD_PTR(PARA_MIXER));
	else
		return;
	printf("%s", help);
	free(help);
	show_subcommands();
	exit(EXIT_SUCCESS);
}

static int parse_and_merge_config_file(const struct lls_command *cmd)
{
	int ret;
	int cf_argc;
	char **cf_argv;
	char *cf, *errctx = NULL;
	struct lls_parse_result **lprp, *cf_lpr, *merged_lpr;
	void *map;
	size_t sz;
	const char *subcmd_name;

	if (cmd == lls_cmd(0, mixer_suite)) {
		lprp = &lpr;
		subcmd_name = NULL;
	} else {
		lprp = &sub_lpr;
		subcmd_name = lls_command_name(cmd);
	}
	if (OPT_GIVEN(PARA_MIXER, CONFIG_FILE))
		cf = para_strdup(OPT_STRING_VAL(PARA_MIXER, CONFIG_FILE));
	else {
		char *home = para_homedir();
		cf = make_message("%s/.paraslash/mixer.conf", home);
		free(home);
	}
	ret = mmap_full_file(cf, O_RDONLY, &map, &sz, NULL);
	if (ret < 0) {
		if (ret != -E_EMPTY && ret != -ERRNO_TO_PARA_ERROR(ENOENT))
			goto free_cf;
		if (ret == -ERRNO_TO_PARA_ERROR(ENOENT) &&
				OPT_GIVEN(PARA_MIXER, CONFIG_FILE))
			goto free_cf;
	} else {
		ret = lls(lls_convert_config(map, sz, subcmd_name, &cf_argv, &errctx));
		para_munmap(map, sz);
		if (ret < 0)
			goto free_cf;
		cf_argc = ret;
		ret = lls(lls_parse(cf_argc, cf_argv, cmd, &cf_lpr, &errctx));
		lls_free_argv(cf_argv);
		if (ret < 0)
			goto free_cf;
		ret = lls(lls_merge(*lprp, cf_lpr, cmd, &merged_lpr, &errctx));
		lls_free_parse_result(cf_lpr, cmd);
		if (ret < 0)
			goto free_cf;
		lls_free_parse_result(*lprp, cmd);
		*lprp = merged_lpr;
		loglevel = OPT_UINT32_VAL(PARA_MIXER, LOGLEVEL);
	}
	ret = 1;
free_cf:
	free(cf);
	if (errctx)
		PARA_ERROR_LOG("%s\n", errctx);
	free(errctx);
	return ret;
}

/**
 * The main function of para_mixer.
 *
 * The executable is linked with the alsa or the oss mixer API, or both. It has
 * a custom log function which prefixes log messages with the current date.
 *
 * \param argc Argument counter.
 * \param argv Argument vector.
 *
 * \return EXIT_SUCCESS or EXIT_FAILURE.
 */
int main(int argc, char *argv[])
{
	const struct lls_command *cmd = CMD_PTR(PARA_MIXER);
	int ret;
	char *errctx;
	const char *subcmd;
	const struct mixer *m;
	struct mixer_handle *h;
	unsigned n;

	ret = lls(lls_parse(argc, argv, cmd, &lpr, &errctx));
	if (ret < 0)
		goto fail;
	loglevel = OPT_UINT32_VAL(PARA_MIXER, LOGLEVEL);
	version_handle_flag("mixer", OPT_GIVEN(PARA_MIXER, VERSION));
	handle_help_flags();

	n = lls_num_inputs(lpr);
	if (n == 0) {
		show_subcommands();
		ret = 0;
		goto free_lpr;
	}
	ret = parse_and_merge_config_file(cmd);
	if (ret < 0)
		goto free_lpr;
	subcmd = lls_input(0, lpr);
	ret = lls(lls_lookup_subcmd(subcmd, mixer_suite, &errctx));
	if (ret < 0)
		goto fail;
	cmd = lls_cmd(ret, mixer_suite);
	ret = lls(lls_parse(n, argv + argc - n, cmd, &sub_lpr, &errctx));
	if (ret < 0)
		goto free_lpr;
	ret = parse_and_merge_config_file(cmd);
	if (ret < 0)
		goto free_lpr;
	m = get_mixer_or_die();
	ret = m->open(OPT_STRING_VAL(PARA_MIXER, MIXER_DEVICE), &h);
	if (ret < 0)
		goto free_sub_lpr;
	ret = set_channel(m, h, OPT_STRING_VAL(PARA_MIXER, MIXER_CHANNEL));
	if (ret == -E_BAD_CHANNEL) {
		char *channels = m->get_channels(h);
		printf("Available channels: %s\n", channels);
		free(channels);
	}
	if (ret < 0)
		goto close_mixer;
	ret = (*(mixer_subcommand_handler_t *)(lls_user_data(cmd)))(m ,h);
close_mixer:
	m->close(&h);
free_sub_lpr:
	lls_free_parse_result(sub_lpr, cmd);
free_lpr:
	lls_free_parse_result(lpr, CMD_PTR(PARA_MIXER));
	if (ret >= 0)
		return EXIT_SUCCESS;
fail:
	if (errctx)
		PARA_ERROR_LOG("%s\n", errctx);
	free(errctx);
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	return EXIT_FAILURE;
}
