/*
 * Copyright (C) 2006 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write_common.c common functions of para_audiod and para_write */

#include <regex.h>
#include <lopsub.h>

#include "write_cmd.lsg.h"
#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "buffer_tree.h"
#include "write.h"
#include "error.h"

/** Loop over all writers. */
#define FOR_EACH_WRITER(i) for (i = 1; lls_cmd(i, write_cmd_suite); i++)


static inline bool writer_supported(int wid)
{
	return lls_user_data(WRITE_CMD(wid));
}

/* simply return the first available writer */
static int default_writer_id(void)
{
	int i;

	FOR_EACH_WRITER(i)
		if (writer_supported(i))
			return i;
	assert(0); /* the file writer should always be available */
}

/**
 * Return the writer structure from a writer ID.
 *
 * \param wid If non-positive, a pointer to the default writer is returned.
 *
 * \return Pointer to a (constant) struct writer.
 */
const struct writer *writer_get(int wid)
{
	if (wid < 0)
		wid = default_writer_id();
	return lls_user_data(WRITE_CMD(wid));
}

/**
 * Return name of the writer identified by a writer ID.
 *
 * \param wid If non-positive, the name of the default writer is returned.
 *
 * \return The returned buffer must not be freed by the caller.
 */
const char *writer_name(int wid)
{
	if (wid <= 0)
		wid = default_writer_id();
	return lls_command_name(WRITE_CMD(wid));
}

/**
 * Check if the given string is a valid command line for any writer.
 *
 * \param wa String of the form writer_name options.
 * \param lprp Contains the parsed command line on success.
 *
 * If wa is \p NULL, the (configuration-dependent) default writer is assumed.
 * Otherwise, the function checks whether \a wa starts with the name of a
 * supported writer. If a valid writer name was found, the rest of the command
 * line is passed to the config parser of this writer.
 *
 * \return On success, the positive writer ID is returned. Otherwise the
 * function prints an error message and calls exit().
 */
int check_writer_arg_or_die(const char *wa, struct lls_parse_result **lprp)
{
	int ret, writer_num, argc;
	char **argv = NULL, *errctx = NULL;
	const struct lls_command *cmd;

	if (!wa || !*wa) {
		writer_num = default_writer_id();
		cmd = WRITE_CMD(writer_num);
		argv = para_malloc(2 * sizeof(char *));
		argc = 1;
		argv[0] = para_strdup(lls_command_name(cmd));
		argv[1] = NULL;
		goto parse;
	}
	ret = create_argv(wa, " \t\n", &argv);
	if (ret < 0)
		goto fail;
	argc = ret;
	ret = lls(lls_lookup_subcmd(argv[0], write_cmd_suite, &errctx));
	if (ret < 0)
		goto free_argv;
	writer_num = ret;
	cmd = WRITE_CMD(writer_num);
	if (!writer_supported(writer_num)) {
		ret = -ERRNO_TO_PARA_ERROR(EINVAL);
		errctx = make_message("%s writer is not supported",
			lls_command_name(cmd));
		goto free_argv;
	}
parse:
	ret = lls(lls_parse(argc, argv, cmd, lprp, &errctx));
	if (ret >= 0)
		ret = writer_num;
free_argv:
	free_argv(argv);
	if (ret >= 0)
		return ret;
fail:
	if (errctx)
		PARA_ERROR_LOG("%s\n", errctx);
	free(errctx);
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

/**
 * Open a writer node and register the corresponding task.
 *
 * \param wn The writer node to open.
 * \param parent The parent btr node (the source for the writer node).
 * \param s The scheduler instance to register the task to.
 */
void register_writer_node(struct writer_node *wn, struct btr_node *parent,
		struct sched *s)
{
	const struct writer *w = writer_get(wn->wid);

	wn->btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = writer_name(wn->wid), .parent = parent,
		.handler = w->execute, .context = wn));
	wn->task = task_register(&(struct task_info) {
		.name = writer_name(wn->wid),
		.pre_select = w->pre_select,
		.post_select = w->post_select,
		.context = wn,
	}, s);
}

/**
 * Print the help text of all writers to stdout.
 *
 * \param detailed Whether to print the short or the detailed help.
 */
void print_writer_helps(bool detailed)
{
	int i;

	printf("\nAvailable writers: ");
	FOR_EACH_WRITER(i) {
		if (!writer_supported(i))
			continue;
		printf("%s%s", i? " " : "", writer_name(i));
	}
	printf("\n");
	FOR_EACH_WRITER(i) {
		const struct lls_command *cmd = WRITE_CMD(i);
		char *help;
		if (!writer_supported(i))
			continue;
		help = detailed? lls_long_help(cmd) : lls_short_help(cmd);
		if (!help)
			continue;
		printf("%s\n", help);
		free(help);
	}
}

static void get_btr_value(struct btr_node *btrn, const char *cmd,
		int32_t *result)
{
	char *buf = NULL;
	int ret = btr_exec_up(btrn, cmd, &buf);

	if (ret < 0) {
		/*
		 * This really should not happen. It means one of our parent
		 * nodes died unexpectedly. Proceed with fingers crossed.
		 */
		PARA_CRIT_LOG("cmd %s: %s\n", cmd, para_strerror(-ret));
		*result = 0;
		return;
	}
	ret = para_atoi32(buf, result);
	assert(ret >= 0);
	free(buf);
}

/**
 * Ask parent btr nodes for the sample rate of the current stream.
 *
 * \param btrn Where to start the search.
 * \param result Filled in by this function.
 *
 * This function is assumed to succeed and terminates on errors.
 */
void get_btr_sample_rate(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "sample_rate", result);
}

/**
 * Ask parent btr nodes for the channel count of the current stream.
 *
 * \param btrn See \ref get_btr_sample_rate.
 * \param result See \ref get_btr_sample_rate.
 */
void get_btr_channels(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "channels", result);
}

/**
 * Ask parent btr nodes for the number of bits per sample and the byte sex.
 *
 * \param btrn See \ref get_btr_sample_rate.
 * \param result Contains the sample format as an enum sample_format type.
 */
void get_btr_sample_format(struct btr_node *btrn, int32_t *result)
{
	get_btr_value(btrn, "sample_format", result);
}
